with Ada.Unchecked_Deallocation;

package body Two_Way_Linked_Lists is

   -- ------------------------------------------------------------------------

   function New_Node (Element : Element_Type) return not null Node_Access is
   begin
      return new Node_Type'
        (Element  => Element,
         Previous => null,
         Next     => null);
   end New_Node;

   -- ------------------------------------------------------------------------

   procedure Free (Node : in out Node_Access) is
      procedure Free_Memory is new Ada.Unchecked_Deallocation
        (Object => Node_Type,
         Name   => Node_Access);
   begin
      Free_Memory (Node);
      -- Set Node to null by the way.
   end Free;

   -- ------------------------------------------------------------------------

   procedure Free (Self : in out Self_Access) is
      procedure Free_Memory is new Ada.Unchecked_Deallocation
        (Object => Self_Type,
         Name   => Self_Access);
   begin
      Free_Memory (Self);
   end Free;

   -- ------------------------------------------------------------------------

   procedure Initialize (List : in out Instance_Type) is
   begin
      List.Self := new Self_Type;
      List.Self.First := null;
      List.Self.Last  := null;
   end Initialize;

   -- ------------------------------------------------------------------------

   procedure Finalize (List : in out Instance_Type) is
      Node_To_Free : Node_Access;
      Next         : Node_Access;
   begin
      Next := List.Self.First;

      while Next /= null loop
         Node_To_Free := Next;
         Next         := Next.Next;
         Free (Node_To_Free);
      end loop;

      List.Self.First := null;
      List.Self.Last  := null;

      Free (List.Self);

   end Finalize;

   -- ------------------------------------------------------------------------

   function Is_Empty (List : Instance_Type) return Boolean is
   begin
      return List.Self.First = null;
   end Is_Empty;

   -- ------------------------------------------------------------------------

   function First (List : Instance_Type) return Iterator_Type
   is
   begin
      return Result : Iterator_Type do
         Result.Self := List.Self;
         Move_To_First (Result);
      end return;
   end First;

   -- ------------------------------------------------------------------------

   function Not_Null (Iterator : Iterator_Type) return Boolean is
   begin
      return Iterator.State = Valid;
   end Not_Null;

   -- ------------------------------------------------------------------------

   procedure Ensure_Is_Not_Null (Iterator : in Iterator_Type) is
   begin
      if Iterator.State /= Valid then
         raise Error;
      end if;
   end Ensure_Is_Not_Null;

   -- ------------------------------------------------------------------------

   procedure Move_To_First (Iterator : in out Iterator_Type) is
      First_Node : constant Node_Access := Iterator.Self.First;
   begin
      Iterator.Node := First_Node;

      if First_Node = null then
         Iterator.State := End_Of_List;
      else
         Iterator.State := Valid;
      end if;
   end Move_To_First;

   -- ------------------------------------------------------------------------

   procedure Move_To_Next (Iterator : in out Iterator_Type) is
   begin
      case Iterator.State is
         when Valid =>
            declare
               Next_Node : constant Node_Access := Iterator.Node.Next;
            begin
               Iterator.Node := Next_Node;
               if Next_Node /= null then
                  Iterator.State := Valid;
               else
                  Iterator.State := End_Of_List;
               end if;
            end;
         when End_Of_List =>
            raise Error;
         when Start_Of_List =>
            Move_To_First (Iterator);
      end case;
   end Move_To_Next;

   -- ------------------------------------------------------------------------

   function Element_At (Iterator : Iterator_Type) return Element_Type is
   begin
      Ensure_Is_Not_Null (Iterator);

      return Iterator.Node.Element;
   end Element_At;

end Two_Way_Linked_Lists;
