private with Ada.Finalization;

generic

   type Element_Type is private;

package Two_Way_Linked_Lists is

   Error : exception;
   type Instance_Type is limited private;
   type Iterator_Type (<>) is limited private;

   -- ------------------------------------------------------------------------

   function Is_Empty (List : Instance_Type) return Boolean;

   -- ------------------------------------------------------------------------

   function Not_Null (Iterator : Iterator_Type) return Boolean;
   function First (List : Instance_Type) return Iterator_Type;
   procedure Move_To_First (Iterator : in out Iterator_Type);
   procedure Move_To_Next (Iterator : in out Iterator_Type);
   function Element_At (Iterator : Iterator_Type) return Element_Type;

private

   type Node_Type;

   type Node_Access is access Node_Type;

   type Node_Type is record
      Element  : Element_Type;
      Previous : Node_Access := null;
      Next     : Node_Access := null;
   end record;

   type Self_Type is record
      First : Node_Access := null;
      Last  : Node_Access := null;
   end record;

   type Self_Access is access Self_Type;

   type Instance_Type is new Ada.Finalization.Limited_Controlled with record
      Self : Self_Access := null;
   end record;

   pragma Warnings (Off);
   overriding procedure Initialize (List : in out Instance_Type);
   overriding procedure Finalize (List : in out Instance_Type);
   pragma Warnings (On);

   type Iterator_State_Kinds is (Valid, Start_Of_List, End_Of_List);

   type Iterator_Type is record
      State : Iterator_State_Kinds;
      Node  : Node_Access;
      Self  : Self_Access := null;
   end record;

end Two_Way_Linked_Lists;
