with Ada.Unchecked_Deallocation;

package body Windows.Environment_Variables is

   -- ------------------------------------------------------------------------


   procedure Free (Item : in out Item_Access)
   is
      procedure Free_Memory is new Ada.Unchecked_Deallocation
        (Object => Item_Type,
         Name   => Item_Access);
   begin
      Free_Memory (Item);
      -- Set Item to null by the way.
   end;

   -- ------------------------------------------------------------------------

   procedure Initialize (Instance : in out Instance_Type) is
   begin
      null;
   end Initialize;

   -- ------------------------------------------------------------------------

   procedure Finalize (Instance : in out Instance_Type) is
      Current_Item     : Item_Access;
      Current_Position : Iterator_Type := First (Instance);
   begin
      while Not_Null (Current_Position) loop
         Current_Item := Item (Current_Position);
         Free (Current_Item);
         Move_To_Next (Current_Position);
      end loop;
   end Finalize;

   -- ------------------------------------------------------------------------

   function Is_Empty (Instance : Instance_Type) return Boolean
   is
   begin
      return Containers.Is_Empty (Instance.List);
   end;

   -- ------------------------------------------------------------------------

   package body Iterators is

      function First (List : Source_List_Type) return Instance_Type
      is
      begin
         return Instance : Instance_Type do
            Instance.Target_Iterator_Reference :=
              (new Target_Iterator_Reference_Type);
            Instance.Target_Iterator_Reference.Target_Iterator :=
              (new Target_Iterator_Type'(Containers.First (List)));
            Instance.Target_Iterator_Reference.Count :=
              (Reference_Count_Type'Succ
                 (Instance.Target_Iterator_Reference.Count));
         end return;
      end;

      -- ---------------------------------------------------------------------

      function Iterator
        (Instance : Instance_Type)
         return access Target_Iterator_Type
      is
      begin
         return Instance.Target_Iterator_Reference.Target_Iterator;
      end;

      -- ---------------------------------------------------------------------

      overriding procedure Adjust (Instance : in out Instance_Type)
      is
      begin
         Instance.Target_Iterator_Reference.Count :=
           (Reference_Count_Type'Succ
              (Instance.Target_Iterator_Reference.Count));
      end;

      -- ---------------------------------------------------------------------

      procedure Free (Target_Iterator : in out Target_Iterator_Access)
      is
         procedure Free_Memory is new Ada.Unchecked_Deallocation
           (Object => Target_Iterator_Type,
            Name   => Target_Iterator_Access);
      begin
         Free_Memory (Target_Iterator);
      end;

      -- ---------------------------------------------------------------------

      procedure Free
        (Target_Iterator_Reference : in out Target_Iterator_Reference_Access)
      is
         procedure Free_Memory is new Ada.Unchecked_Deallocation
           (Object => Target_Iterator_Reference_Type,
            Name   => Target_Iterator_Reference_Access);
      begin
         Free_Memory (Target_Iterator_Reference);
      end;

      -- ---------------------------------------------------------------------

      overriding procedure Finalize (Instance : in out Instance_Type) is
      begin
         Instance.Target_Iterator_Reference.Count :=
           (Reference_Count_Type'Pred
              (Instance.Target_Iterator_Reference.Count));
         if
           ((Instance.Target_Iterator_Reference.Count) =
            (Reference_Count_Type'First))
         then
            Free (Instance.Target_Iterator_Reference.Target_Iterator);
            Free (Instance.Target_Iterator_Reference);
         end if;
      end;

   end Iterators;

   -- ------------------------------------------------------------------------

   function Conversion
     (Source : Iterator_Type)
      return access Containers.Iterator_Type
   is
   begin
      return Iterators.Iterator (Iterators.Instance_Type (Source));
   end;

   -- ------------------------------------------------------------------------

   function Not_Null (Iterator : Iterator_Type) return Boolean
   is
   begin
      return Containers.Not_Null (Conversion (Iterator).all);
   end;

   -- ------------------------------------------------------------------------

   function First (Instance : Instance_Type) return Iterator_Type
   is
   begin
      return Iterator_Type (Iterators.First (Instance.List));
   end;

   -- ------------------------------------------------------------------------

   procedure Move_To_Next (Iterator : in out Iterator_Type)
   is
   begin
      Containers.Move_To_Next (Conversion (Iterator).all);
   end;

   -- ------------------------------------------------------------------------

   function Item (Iterator : Iterator_Type) return Item_Access
   is
   begin
      return Containers.Element_At (Conversion (Iterator).all);
   end;

   -- ------------------------------------------------------------------------

   function Name (Iterator : Iterator_Type) return String
   is
   begin
      return String (Item (Iterator).Name);
   end;

   -- ------------------------------------------------------------------------

   function Value (Iterator : Iterator_Type) return Value_Type
   is
   begin
      return Item (Iterator).Value;
   end;

end Windows.Environment_Variables;
