private with Two_Way_Linked_Lists;
private with Ada.Finalization;

package Windows.Environment_Variables is

   Error : exception;

   -- ------------------------------------------------------------------------

   type Name_Type  is new String;
   subtype Value_Type is String;

   -- ------------------------------------------------------------------------

   type Instance_Type is limited private;
   function Is_Empty (Instance : Instance_Type) return Boolean;

   -- ------------------------------------------------------------------------

   type Iterator_Type (<>) is limited private;
   function Not_Null (Iterator : Iterator_Type) return Boolean;
   function First (Instance : Instance_Type) return Iterator_Type;
   procedure Move_To_Next (Iterator : in out Iterator_Type);
   function Name (Iterator : Iterator_Type) return String;
   function Value (Iterator : Iterator_Type) return Value_Type;

private

   type Item_Type (Name_Length : Natural; Value_Length : Natural) is record
      Name  : Name_Type (1 .. Name_Length);
      Value : Value_Type (1 .. Value_Length);
   end record;

   type Item_Access is access Item_Type;

   -- ------------------------------------------------------------------------

   package Containers is new Two_Way_Linked_Lists
     (Element_Type => Item_Access);

   -- ------------------------------------------------------------------------

   type Instance_Type is new Ada.Finalization.Limited_Controlled with record
      List : Containers.Instance_Type;
   end record;

   overriding procedure Initialize (Instance : in out Instance_Type);
   overriding procedure Finalize (Instance : in out Instance_Type);

   -- ------------------------------------------------------------------------

   package Iterators is

      subtype Source_List_Type is Containers.Instance_Type;
      subtype Target_Iterator_Type is Containers.Iterator_Type;

      type Instance_Type (<>) is private;

      function First (List : Source_List_Type) return Instance_Type;

      function Iterator
        (Instance : Instance_Type)
         return access Target_Iterator_Type;

   private

      type Target_Iterator_Access is access Target_Iterator_Type;

      type Reference_Count_Type is new Natural;

      type Target_Iterator_Reference_Type is record
         Count           : Reference_Count_Type := Reference_Count_Type'First;
         Target_Iterator : Target_Iterator_Access := null;
      end record;

      type Target_Iterator_Reference_Access is
        access Target_Iterator_Reference_Type;

      type Instance_Type is new Ada.Finalization.Controlled
      with record
         Target_Iterator_Reference : Target_Iterator_Reference_Access :=
           (null);
      end record;

      overriding procedure Adjust   (Instance : in out Instance_Type);
      overriding procedure Finalize (Instance : in out Instance_Type);

   end Iterators;

   -- ------------------------------------------------------------------------

   type Iterator_Type is new Iterators.Instance_Type;

   -- ------------------------------------------------------------------------

   function Item (Iterator : Iterator_Type) return Item_Access;

end Windows.Environment_Variables;
