with Windows.Threads;
with Windows.Startup_Informations;
with Windows.Security_Attributes;
with Windows.Environment_Variables;

package Windows.Processes is


   pragma Elaborate_Body;

   Error : exception;

   -- ------------------------------------------------------------------------

   subtype Thread_Type is Threads.Thread_Type;
   subtype Thread_Identifier_Type is Threads.Thread_Identifier_Type;

   -- ------------------------------------------------------------------------

   type Process_Informations_Type is limited private;
   type Process_Type              is limited private;
   type Process_Identifier_Type   is limited private;

   -- ------------------------------------------------------------------------

   subtype Security_Attributes_Type is Security_Attributes.Instance_Type;

   -- ------------------------------------------------------------------------

   subtype Startup_Informations_Type is Startup_Informations.Instance_Type;

   -- ------------------------------------------------------------------------

   type Startup_Options_Type is limited private;

   -- ------------------------------------------------------------------------

   function Run
     (Command_Line         : String;
      Startup_Directory    : String;
      Startup_Informations : Startup_Informations_Type;
      Startup_Options      : Startup_Options_Type)
      return Process_Informations_Type;

   -- ------------------------------------------------------------------------

   type Exit_Code_Type is new Unsigned_Type;

   -- ------------------------------------------------------------------------

   function Run
     (Command_Line         : String;
      Startup_Informations : Startup_Informations_Type;
      Startup_Options      : Startup_Options_Type)
      return Exit_Code_Type;

   -- ------------------------------------------------------------------------

   function Run
     (Command_Line         : String;
      Startup_Directory    : String;
      Startup_Informations : Startup_Informations_Type;
      Startup_Options      : Startup_Options_Type)
      return Exit_Code_Type;

private

   type Process_Type is new Handle_Type; -- !!! Make Handle_Type limited ?
   Null_Process_Handle : constant Process_Type := Process_Type (Null_Handle);

   -- ------------------------------------------------------------------------

   type Process_Identifier_Type is new Identifier_Type;

   Null_Process_Identifier : constant Process_Identifier_Type :=
     (Process_Identifier_Type (Null_Identifier));

   -- ------------------------------------------------------------------------

   type Process_Informations_Type is record
      Process : Process_Type := Null_Process_Handle;
      Primary_Thread : Thread_Type := Threads.No_Thread;
      Process_Identifier : Process_Identifier_Type := Null_Process_Identifier;
      Primary_Thread_Identifier : Thread_Identifier_Type := Threads.No_Thread;
   end record;

   pragma Pack (Process_Informations_Type);
   pragma Convention
     (Convention => C_Pass_By_Copy,
      Entity     => Process_Informations_Type);

   -- ------------------------------------------------------------------------

   type Creation_Mode_Type is record
      Debug_This_Process_And_Child_Processes : Bit_Boolean_Type := False;
      Debug_Only_This_Process                : Bit_Boolean_Type := False;
      Start_Suspended                        : Bit_Boolean_Type := False;
      Do_Not_Inherit_The_Console             : Bit_Boolean_Type := False;
      Get_A_New_Console                      : Bit_Boolean_Type := False;
      Padding_1                              : Padding_Type := Padding_Value;
      As_Root_Of_A_New_Process_Group         : Bit_Boolean_Type := False;
      Environment_Variables_Are_Unicode      : Bit_Boolean_Type := False;
      With_Its_Own_Virtual_DOS_Machine       : Bit_Boolean_Type := False;
      Use_The_Shared_Virtual_DOS_Machine     : Bit_Boolean_Type := False;
      Undocumented_Force_DOS                 : Bit_Boolean_Type := False;
      Padding_2                              : Padding_Type := Padding_Value;
      Inherit_Parent_Processor_Affinity      : Bit_Boolean_Type := False;
      Padding_3                              : Padding_Type := Padding_Value;
      As_Protected_Process                   : Bit_Boolean_Type := False;
      Use_Extended_Startup_Informations      : Bit_Boolean_Type := False;
      Padding_4                              : Padding_Type := Padding_Value;
      Not_Associated_To_My_Job               : Bit_Boolean_Type := False;
      Bypass_Normal_Process_Restrictions     : Bit_Boolean_Type := False;
      Do_No_Inherit_Error_Mode               : Bit_Boolean_Type := False;
      With_No_Console_And_No_GUI             : Bit_Boolean_Type := False;
      Padding_5                              : Padding_Type := Padding_Value;
   end record;

   for Creation_Mode_Type'Size use Unsigned_Type'Size;
   pragma Convention
     (Convention => C_Pass_By_Copy,
      Entity     => Creation_Mode_Type);

   for Creation_Mode_Type use record
      Debug_This_Process_And_Child_Processes at 0 range 0 .. 0;
      Debug_Only_This_Process                at 0 range 1 .. 1;
      Start_Suspended                        at 0 range 2 .. 2;
      Do_Not_Inherit_The_Console             at 0 range 3 .. 3;
      Get_A_New_Console                      at 0 range 4 .. 4;
      Padding_1                              at 0 range 5 .. 8;
      As_Root_Of_A_New_Process_Group         at 0 range 9 .. 9;
      Environment_Variables_Are_Unicode      at 0 range 10 .. 10;
      With_Its_Own_Virtual_DOS_Machine       at 0 range 11 .. 11;
      Use_The_Shared_Virtual_DOS_Machine     at 0 range 12 .. 12;
      Undocumented_Force_DOS                 at 0 range 13 .. 13;
      Padding_2                              at 0 range 14 .. 15;
      Inherit_Parent_Processor_Affinity      at 0 range 16 .. 16;
      Padding_3                              at 0 range 17 .. 17;
      As_Protected_Process                   at 0 range 18 .. 18;
      Use_Extended_Startup_Informations      at 0 range 19 .. 19;
      Padding_4                              at 0 range 20 .. 23;
      Not_Associated_To_My_Job               at 0 range 24 .. 24;
      Bypass_Normal_Process_Restrictions     at 0 range 25 .. 25;
      Do_No_Inherit_Error_Mode               at 0 range 26 .. 26;
      With_No_Console_And_No_GUI             at 0 range 27 .. 27;
      Padding_5                              at 0 range 28 .. 31;
   end record;

   -- ------------------------------------------------------------------------

   subtype Environment_Variables_Type is
     Windows.Environment_Variables.Instance_Type;

   type Startup_Options_Type is limited record
      Creation_Mode                              : Creation_Mode_Type;
      Grant_Access_To_Inheritable_Objects        : Boolean := True;
      Process_Security_Attributes_Defined        : Boolean := False;
      Process_Security_Attributes                : Security_Attributes_Type;
      Primary_Thread_Security_Attributes_Defined : Boolean := False;
      Primary_Thread_Security_Attributes         : Security_Attributes_Type;
      Environment_Variables_Defined              : Boolean := False;
      Environment_Variables                      : Environment_Variables_Type;
   end record;

end Windows.Processes;
