with Windows.Streams;
with Windows.Widgets;
private with Windows.Streams.Clones;

package Windows.Startup_Informations is

   -- ------------------------------------------------------------------------

   type Instance_Type is limited private;

   -- ------------------------------------------------------------------------

   subtype Input_Type  is Streams.Input_Type;
   subtype Output_Type is Streams.Output_Type;

   -- ------------------------------------------------------------------------

   subtype Display_Mode_Type is Widgets.Display_Mode_Type;

   -- ------------------------------------------------------------------------

   type Character_Attributes_Type is private;

   Default_Character_Attributes : constant Character_Attributes_Type;

private

   -- ------------------------------------------------------------------------

   type Character_Attributes_Type is record
      Foreground_Contains_Blue  : Bit_Boolean_Type := False;
      Foreground_Contains_Green : Bit_Boolean_Type := False;
      Foreground_Contains_Red   : Bit_Boolean_Type := False;
      Foreground_Is_Intensified : Bit_Boolean_Type := False;
      Background_Contains_Blue  : Bit_Boolean_Type := False;
      Background_Contains_Green : Bit_Boolean_Type := False;
      Background_Contains_Red   : Bit_Boolean_Type := False;
      Background_Is_Intensified : Bit_Boolean_Type := False;
   end record;

   for Character_Attributes_Type use record
      Foreground_Contains_Blue    at 0 range 0 .. 0;
        Foreground_Contains_Green at 0 range 1 .. 1;
        Foreground_Contains_Red   at 0 range 2 .. 2;
        Foreground_Is_Intensified at 0 range 3 .. 3;
        Background_Contains_Blue  at 0 range 4 .. 4;
        Background_Contains_Green at 0 range 5 .. 5;
        Background_Contains_Red   at 0 range 6 .. 6;
        Background_Is_Intensified at 0 range 7 .. 7;
   end record;

   Character_Attributes_Type_Size : constant := 8;

   for Character_Attributes_Type'Size use Character_Attributes_Type_Size;
   pragma Convention
     (Convention => C_Pass_By_Copy,
      Entity     => Character_Attributes_Type);

   -- ------------------------------------------------------------------------

   Default_Character_Attributes : constant Character_Attributes_Type :=
     (Foreground_Contains_Blue  => True,
      Foreground_Contains_Green => True,
      Foreground_Contains_Red   => True,
      Foreground_Is_Intensified => False,
      Background_Contains_Blue  => False,
      Background_Contains_Green => False,
      Background_Contains_Red   => False,
      Background_Is_Intensified => False);

   -- ------------------------------------------------------------------------

   type Options_Type is record
      GUI_Window_Default_Display_Mode_Is_Defined : Bit_Boolean_Type := False;
      GUI_Window_Default_Size_Is_Defined         : Bit_Boolean_Type := False;
      GUI_Window_Default_Position_Is_Defined     : Bit_Boolean_Type := False;
      Console_Size_Is_Defined                    : Bit_Boolean_Type := False;
      Console_Character_Attributes_Is_Defined    : Bit_Boolean_Type := False;
      Run_In_Full_Scree_Mode                     : Bit_Boolean_Type := False;
      Use_Feedback_Cursor                        : Bit_Boolean_Type := False;
      Do_Not_Use_Feedback_Cursor                 : Bit_Boolean_Type := False;
      Standard_Streams_Are_Defined               : Bit_Boolean_Type := False;
      Undocumented_Use_Hotkey                    : Bit_Boolean_Type := False;
      Padding                                : Padding_Type := Padding_Value;
   end record;

   for Options_Type use record
      GUI_Window_Default_Display_Mode_Is_Defined at 0 range 0 .. 0;
      GUI_Window_Default_Size_Is_Defined         at 0 range 1 .. 1;
      GUI_Window_Default_Position_Is_Defined     at 0 range 2 .. 2;
      Console_Size_Is_Defined                    at 0 range 3 .. 3;
      Console_Character_Attributes_Is_Defined    at 0 range 4 .. 4;
      Run_In_Full_Scree_Mode                     at 0 range 5 .. 5;
      Use_Feedback_Cursor                        at 0 range 6 .. 6;
      Do_Not_Use_Feedback_Cursor                 at 0 range 7 .. 7;
      Standard_Streams_Are_Defined               at 0 range 8 .. 8;
      Undocumented_Use_Hotkey                    at 0 range 9 .. 9;
      Padding at 0 range 10 .. Unsigned_Type'Size - 1;
   end record;

   for Options_Type'Size use Unsigned_Type'Size;
   pragma Convention (Convention => C_Pass_By_Copy, Entity => Options_Type);

   -- ------------------------------------------------------------------------

   type Reserved_1_Type is new Reserved_Type;
   for Reserved_1_Type'Size use Access_Type'Size;

   type Reserved_2_Type is new Reserved_Type;
   for Reserved_2_Type'Size use Small_Unsigned_Type'Size;

   type Reserved_3_Type is new Reserved_Type;
   for Reserved_3_Type'Size use Access_Type'Size;

   type Character_Attributes_Padding_Type is new Padding_Type;
   for Character_Attributes_Padding_Type'Size use
     (Unsigned_Type'Size - Character_Attributes_Type_Size);

   -- ------------------------------------------------------------------------

   type Instance_Type is limited record
      Record_Size_In_Byte : Unsigned_Type := (Instance_Type'Size / Byte_Size);
      Reserved_1 : Reserved_1_Type := Reserved_Value;
      Desktop_Name : Access_Type := Null_Access;
      Console_Title : Access_Type := Null_Access;
      GUI_Window_Default_Position_X : Unsigned_Type := 0;
      GUI_Window_Default_Position_Y : Unsigned_Type := 0;
      GUI_Window_Default_Width : Unsigned_Type := 0;
      GUI_Window_Default_Height : Unsigned_Type := 0;
      Console_Width : Unsigned_Type := 0;
      Console_Height : Unsigned_Type := 0;
      Console_Character_Attributes  : Character_Attributes_Type :=
        (Default_Character_Attributes);
      Character_Attribute_Padding   : Character_Attributes_Padding_Type :=
        (Padding_Value);
      Options : Options_Type;
      GUI_Window_Default_Display_Mode : Display_Mode_Type :=
        (Widgets.Show_As_For_The_First_Time);
      Reserved_2 : Reserved_2_Type := Reserved_Value;
      Reserved_3 : Reserved_3_Type := Reserved_Value;
      Standard_Input : Streams.Clones.Input_Type;
      Standard_Output : Streams.Clones.Output_Type;
      Standard_Error : Streams.Clones.Output_Type;
   end record;

   pragma Pack (Instance_Type);
   pragma Convention (Convention => C_Pass_By_Copy, Entity => Instance_Type);
   pragma Assert ((Instance_Type'Size mod Byte_Size) = 0);

end Windows.Startup_Informations;
