private with Ada.Finalization;
with Windows.Environment_Variables;

package Windows.Strings is


   pragma Elaborate_Body;

   -- ------------------------------------------------------------------------

   Error : exception;

   -- ------------------------------------------------------------------------

   type API_String_Type is tagged limited private;
   type Environment_Block_Type is tagged limited private;

   -- ------------------------------------------------------------------------

   type API_String_Access is limited private;
   type Environment_Block_Access is limited private;

   function Access_To
     (API_String : API_String_Type)
      return       API_String_Access;

   function Access_To_Or_Null
     (API_String     : API_String_Type;
      Null_Condition : Boolean)
      return       API_String_Access;

   function Access_To
     (Environment_Block : Environment_Block_Type)
      return              Environment_Block_Access;

   function Access_To_Or_Null
     (Environment_Block : Environment_Block_Type;
      Null_Condition    : Boolean)
      return              Environment_Block_Access;

   -- ------------------------------------------------------------------------

   Maximum_Length : constant := 8192 - 1;

   function Create
     (Source        : String;
      Null_If_Empty : Boolean := False)
      return API_String_Type;

   function Create
     (Source        : Environment_Variables.Instance_Type;
      Null_If_Empty : Boolean := False)
      return   Environment_Block_Type;

   function Null_String return API_String_Type;

   function Null_String return API_String_Access;

   function Null_Environment_Block return Environment_Block_Type;

   function Null_Environment_Block return Environment_Block_Access;

   -- ------------------------------------------------------------------------

   type Storage_Units_Count_Type is new Signed_Type range
      0 .. (Maximum_Length * 2) + 1;

   for Storage_Units_Count_Type'Size use Signed_Type'Size;

   subtype Data_Length_Type is Storage_Units_Count_Type range
      0 .. Storage_Units_Count_Type'Last - 1;

   function Data_Length
     (API_String : API_String_Type)
      return       Data_Length_Type;

private

   package Finalization renames Ada.Finalization;

   subtype Data_Index_Type is Storage_Units_Count_Type range
      1 .. Storage_Units_Count_Type'Last;

   type String_Data_Type is
     array (Data_Index_Type range <>) of aliased API_Character_Type;

   pragma Convention (Convention => C, Entity => String_Data_Type);

   type API_String_Access is access String_Data_Type;
   for API_String_Access'Size use Access_Type'Size;
   pragma Convention (C, Entity => API_String_Access);

   -- ------------------------------------------------------------------------

   type API_String_Type is new Finalization.Limited_Controlled with record
      String_Data : API_String_Access;
      Data_Length : Data_Length_Type;
   end record;

   pragma Warnings (Off);

   procedure Initialize (S : in out API_String_Type);

   procedure Finalize (S : in out API_String_Type);

   pragma Warnings (On);

   -- ------------------------------------------------------------------------

   type Environment_Block_Access is new API_String_Access;

   type Environment_Block_Type is new API_String_Type with null record;

end Windows.Strings;
