package body Windows is

   -- ------------------------------------------------------------------------

   function Last_API_Error return Error_Code_Type;

   pragma Import
     (Convention => Stdcall,
      Entity => Last_API_Error,
      External_Name => "GetLastError");

   function Last_Error return Error_Code_Type is
   begin
      return Last_API_Error;
   end Last_Error;

   function Last_Error return String is
   begin
      return "Error code #" & Error_Code_Type'Image (Last_Error);
   end Last_Error;

   -- ------------------------------------------------------------------------

   function From_Boolean (Status : Boolean) return Bit_Boolean_Type
   is
   begin
      if Status = True then
         return True;
      else
         return False;
      end if;
   end From_Boolean;

   -- ------------------------------------------------------------------------

   function From_Boolean (Status : Boolean) return Boolean_Type
   is
   begin
      if Status = True then
         return Boolean_Type'Last;
      else
         return Boolean_False;
      end if;
   end From_Boolean;

   -- ------------------------------------------------------------------------

   function From_Boolean (Status : Boolean) return Tiny_Boolean_Type
   is
   begin
      if Status = True then
         return Tiny_Boolean_Type'Last;
      else
         return Tiny_Boolean_False;
      end if;
   end From_Boolean;

   -- ------------------------------------------------------------------------

   procedure Reset (Item : out Options_Type) is
   begin
      Item := 0;
   end Reset;

   procedure Combine
     (Destination : in out Options_Type;
      Options     : in Options_Type)
   is
   begin
      Destination := Destination or Options;
   end Combine;

   -- ------------------------------------------------------------------------

   package body Behaviours is

      type Target_Kinds is (No_Target, By_Object, By_Identifier);

      -- ---------------------------------------------------------------------

      type Behaviour_Target_Type (Target_Kind : Target_Kinds) is record
         case Target_Kind is
            when No_Target =>
               null;
            when By_Object =>
               Object : Messages.Message_Target_Type;
            when By_Identifier =>
               Identifier : Small_Identifier_Type;
         end case;
      end record;

      -- ---------------------------------------------------------------------

      type Behaviour_Implementation_Kinds is (By_Agent, By_Method);

      -- ---------------------------------------------------------------------

      type Association_Type
        (Implementation_Kind : Behaviour_Implementation_Kinds;
         Target_Kind         : Target_Kinds)
      is
         record
            Message_Identifier : Message_Identifier_Type;
            Next               : Association_Access;
            Target             : Behaviour_Target_Type (Target_Kind);

            case Implementation_Kind is
               when By_Agent =>
                  Agent : Agent_Access;
               when By_Method =>
                  Method : Method_Type;
            end case;

         end record;

      -- ---------------------------------------------------------------------

      procedure Execute (Agent : in Agent_Type) is
      begin
         raise Program_Error;
      end Execute;

      -- ---------------------------------------------------------------------

      function New_Behaviours_Map return Behaviours_Map_Type is
      begin
         return Result : Behaviours_Map_Type do
            Result.First := null;
         end return;
      end New_Behaviours_Map;

      -- ---------------------------------------------------------------------

      procedure Insert_At_The_Begining
        (Map      : in out Behaviours_Map_Type;
         New_Node : in Association_Access)
      is
      begin
         if Map.First = null then
            Map.First := New_Node;
         else
            New_Node.Next := Map.First;
            Map.First     := New_Node;
         end if;
      end Insert_At_The_Begining;

      -- =====================================================================

      procedure Associate
        (Map                : in out Behaviours_Map_Type;
         Agent              : in Agent_Access;
         Message_Identifier : in Messages.Message_Identifier_Type)
      is
         New_Node : constant Association_Access :=
            new Association_Type (By_Agent, No_Target);
      begin
         New_Node.Message_Identifier := Message_Identifier;
         New_Node.Agent              := Agent;
         Insert_At_The_Begining (Map, New_Node);
      end Associate;

      -- ---------------------------------------------------------------------

      procedure Associate
        (Map                : in out Behaviours_Map_Type;
         Method             : in Method_Type;
         Message_Identifier : in Messages.Message_Identifier_Type)
      is
         New_Node : constant Association_Access :=
            new Association_Type (By_Method, No_Target);
      begin
         New_Node.Message_Identifier := Message_Identifier;
         New_Node.Method             := Method;
         Insert_At_The_Begining (Map, New_Node);
      end Associate;

      -- =====================================================================

      procedure Associate
        (Map                : in out Behaviours_Map_Type;
         Agent              : in Agent_Access;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Target_Object      : in Messages.Message_Target_Type)
      is
         New_Node : constant Association_Access :=
            new Association_Type (By_Agent, By_Identifier);
      begin
         New_Node.Message_Identifier := Message_Identifier;
         New_Node.Agent              := Agent;
         New_Node.Target.Object      := Target_Object;
         Insert_At_The_Begining (Map, New_Node);
      end Associate;

      -- ---------------------------------------------------------------------

      procedure Associate
        (Map                : in out Behaviours_Map_Type;
         Method             : in Method_Type;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Target_Object      : in Messages.Message_Target_Type)
      is
         New_Node : constant Association_Access :=
            new Association_Type (By_Method, By_Identifier);
      begin
         New_Node.Message_Identifier := Message_Identifier;
         New_Node.Method             := Method;
         New_Node.Target.Object      := Target_Object;
         Insert_At_The_Begining (Map, New_Node);
      end Associate;

      -- =====================================================================

      procedure Associate
        (Map                : in out Behaviours_Map_Type;
         Agent              : in Agent_Access;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Target_Identifier  : in Small_Identifier_Type)
      is
         New_Node : constant Association_Access :=
            new Association_Type (By_Agent, By_Identifier);
      begin
         New_Node.Message_Identifier := Message_Identifier;
         New_Node.Agent              := Agent;
         New_Node.Target.Identifier  := Target_Identifier;
         Insert_At_The_Begining (Map, New_Node);
      end Associate;

      -- ---------------------------------------------------------------------

      procedure Associate
        (Map                : in out Behaviours_Map_Type;
         Method             : in Method_Type;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Target_Identifier  : in Small_Identifier_Type)
      is
         New_Node : constant Association_Access :=
            new Association_Type (By_Method, By_Identifier);
      begin
         New_Node.Message_Identifier := Message_Identifier;
         New_Node.Method             := Method;
         New_Node.Target.Identifier  := Target_Identifier;
         Insert_At_The_Begining (Map, New_Node);
      end Associate;

      -- =====================================================================

      procedure Execute
        (Map                : in out Behaviours_Map_Type;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Parameter_1        : in Variant_Type;
         Parameter_2        : in Variant_Type)
      is
      begin
         Execute
           (Map                => Map,
            Message_Identifier => Message_Identifier,
            Target_Object      => Messages.No_Target,
            Target_Identifier  => Windows.No_Small_Identifier,
            Parameter_1        => Parameter_1,
            Parameter_2        => Parameter_2);
      end Execute;

      -- ---------------------------------------------------------------------

      procedure Execute
        (Map                : in out Behaviours_Map_Type;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Target_Object      : in Messages.Message_Target_Type;
         Parameter_1        : in Variant_Type;
         Parameter_2        : in Variant_Type)
      is
      begin
         Execute
           (Map                => Map,
            Message_Identifier => Message_Identifier,
            Target_Object      => Target_Object,
            Target_Identifier  => Windows.No_Small_Identifier,
            Parameter_1        => Parameter_1,
            Parameter_2        => Parameter_2);
      end Execute;

      -- ---------------------------------------------------------------------

      procedure Execute
        (Map                : in out Behaviours_Map_Type;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Target_Identifier  : in Small_Identifier_Type;
         Parameter_1        : in Variant_Type;
         Parameter_2        : in Variant_Type)
      is
      begin
         Execute
           (Map                => Map,
            Message_Identifier => Message_Identifier,
            Target_Object      => Messages.No_Target,
            Target_Identifier  => Target_Identifier,
            Parameter_1        => Parameter_1,
            Parameter_2        => Parameter_2);
      end Execute;

      -- ---------------------------------------------------------------------

      procedure Execute
        (Map                : in out Behaviours_Map_Type;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Target_Object      : in Messages.Message_Target_Type;
         Target_Identifier  : in Small_Identifier_Type;
         Parameter_1        : in Variant_Type;
         Parameter_2        : in Variant_Type)
      is
         Node          : Association_Access;
         Previous_Node : Association_Access;
         use type Message_Identifier_Type;

         function Node_Match return Boolean is
         begin
            if Node.Message_Identifier /= Message_Identifier then
               return False;
            end if;

            case Node.Target_Kind is
               when No_Target =>
                  return True;
               when By_Object =>
                  return Node.Target.Object = Target_Object;
               when By_Identifier =>
                  return Node.Target.Identifier = Target_Identifier;
            end case;
         end Node_Match;

         procedure Apply_Behaviour is
         begin
            case Node.Implementation_Kind is
               when By_Agent =>
                  Execute (Node.Agent.all);
               when By_Method =>
                  declare
                     Event_Data : constant Event_Data_Type :=
                       (Message_Identifier,
                        Target_Object,
                        Parameter_1,
                        Parameter_2);
                  begin
                     Node.Method.all (Event_Data);
                  end;
            end case;
         end Apply_Behaviour;

         procedure Move_Node_To_Start is
         begin
            if Previous_Node = null then
               -- Already the first one.
               return;
            end if;

            Previous_Node.Next := Node.Next;
            Node.Next          := Map.First;
            Map.First          := Node;
         end Move_Node_To_Start;
      begin
         Previous_Node := null;
         Node          := Map.First;

         while Node /= null loop

            if Node_Match then
               Apply_Behaviour;
               Move_Node_To_Start;
               exit;
            end if;

            Previous_Node := Node;
            Node          := Node.Next;
         end loop;

      end Execute;

   end Behaviours;

   -- ------------------------------------------------------------------------

   package body Agents is

      function New_Agent (Parameters : Parameters_Type) return Agent_Type is
      begin
         return Agent_Type'(Parameters => Parameters);
      end New_Agent;

      overriding procedure Execute (Agent : in Agent_Type) is
      begin
         Execute (Agent.Parameters);
      end Execute;

   end Agents;

   -- ------------------------------------------------------------------------

   package body Parameterless_Agents is

      function New_Agent (Method : Method_Access) return Agent_Type is
      begin
         return Agent_Type'(Method => Method);
      end New_Agent;

      overriding procedure Execute (Agent : in Agent_Type) is
      begin
         Agent.Method.all;
      end Execute;

   end Parameterless_Agents;

end Windows;
