package Windows is

   pragma Elaborate_Body;

   type Unsigned_Type is range 0 .. 4_294_967_295;

   type Signed_Type is range -2_147_483_648 .. +2_147_483_647;

   type Small_Identifier_Type is mod 65536;

   -- ------------------------------------------------------------------------

   Maximum_Path_Length : constant := 260;

   subtype Path_Length_Type is Natural range 0 .. Maximum_Path_Length;

   subtype Path_Text_Index_Type is Path_Length_Type range
     1 .. (Path_Length_Type'Last);

   type Path_Type (Length : Path_Length_Type) is record
      Text : String (Path_Text_Index_Type'First .. Length);
   end record;

   -- ------------------------------------------------------------------------

   type Error_Code_Type is mod 16#1_0000_0000#;

   function Last_Error return Error_Code_Type;

   function Last_Error return String;

   No_Error : constant Error_Code_Type;

   type Exit_Code_Type is range 0 .. 32767;

   Terminates_Successfully : constant Exit_Code_Type;

   -- ------------------------------------------------------------------------

   type GUID_String_Type is new String (1 .. 38);

private

   pragma Linker_Options ("winmain.o");
   pragma Linker_Options ("-lkernel32");
   pragma Linker_Options ("-luser32");
   pragma Linker_Options ("-lgdi32");
   pragma Linker_Options ("-lcomctl32");
   pragma Linker_Options ("-lshlwapi");

   -- ------------------------------------------------------------------------

   Bit_Size : constant := 1;
   Byte_Size : constant := 8;
   Tiny_Size : constant := 8;
   Small_Size : constant := 16;
   Medium_Size : constant := 32;
   Big_Size : constant := 32;

   -- ------------------------------------------------------------------------

   Smallest_Tiny : constant := 0;
   Biggest_Tiny  : constant := (2 ** Tiny_Size) - 1;

   type Tiny_Type is range Smallest_Tiny .. Biggest_Tiny;
   for Tiny_Type'Size use Tiny_Size;
   pragma Convention (Convention => C, Entity => Tiny_Type);

   -- ------------------------------------------------------------------------

   Smallest_Small_Unsigned : constant := 0;
   Biggest_Small_Unsigned  : constant := (2 ** Small_Size) - 1;

   type Small_Unsigned_Type is range
      Smallest_Small_Unsigned .. Biggest_Small_Unsigned;

   for Small_Unsigned_Type'Size use Small_Size;
   pragma Convention (Convention => C, Entity => Small_Unsigned_Type);

   -- ------------------------------------------------------------------------

   Smallest_Small_Signed : constant := -(2 ** (Small_Size - 1));
   Biggest_Small_Signed  : constant := +((2 ** (Small_Size - 1)) - 1);

   type Small_Signed_Type is range
      Smallest_Small_Signed .. Biggest_Small_Signed;

   for Small_Signed_Type'Size use Small_Size;
   pragma Convention (Convention => C, Entity => Small_Signed_Type);

   type Small_Natural_Type is new Small_Signed_Type range
      0 .. Small_Signed_Type'Last;

   -- ------------------------------------------------------------------------

   for Unsigned_Type'Size use Medium_Size;
   pragma Convention (Convention => C, Entity => Unsigned_Type);

   for Signed_Type'Size use Medium_Size;
   pragma Convention (Convention => C, Entity => Signed_Type);

   -- ------------------------------------------------------------------------

   type Natural_Type is new Signed_Type range 0 .. Signed_Type'Last;
   for Natural_Type'Size use Signed_Type'Size;
   pragma Convention (Convention => C, Entity => Natural_Type);

   -- ------------------------------------------------------------------------

   type Bit_Boolean_Type is (False, True);
   for Bit_Boolean_Type'Size use Bit_Size;
   for Bit_Boolean_Type use (False => 0, True => 1);
   function From_Boolean (Status : Boolean) return Bit_Boolean_Type;
   type Boolean_Type is new Signed_Type;
   function From_Boolean (Status : Boolean) return Boolean_Type;
   Boolean_False : constant Boolean_Type := 0;
   type Tiny_Boolean_Type is new Tiny_Type;
   function From_Boolean (Status : Boolean) return Tiny_Boolean_Type;
   Tiny_Boolean_False : constant Tiny_Boolean_Type := 0;
   pragma Inline (From_Boolean);

   -- ------------------------------------------------------------------------

   type Options_Type is mod 2 ** Unsigned_Type'Size;
   for Options_Type'Size use Unsigned_Type'Size;
   pragma Convention (Convention => C, Entity => Options_Type);

   No_Options : constant Options_Type := 0;

   procedure Reset (Item : out Options_Type);

   procedure Combine
     (Destination : in out Options_Type;
      Options     : in Options_Type);

   -- ------------------------------------------------------------------------

   type API_Character_Type is new Wide_Character;
   for API_Character_Type'Size use 16;
   pragma Convention (Convention => C, Entity => API_Character_Type);

   Null_Terminating_Character : constant API_Character_Type :=
      API_Character_Type'Val (0);

   -- ------------------------------------------------------------------------

   type Access_Type is mod 2 ** Big_Size;
   for Access_Type'Size use Big_Size;
   Null_Access : constant Access_Type := 0;

   -- ------------------------------------------------------------------------

   for Small_Identifier_Type'Size use Small_Size;
   pragma Convention (Convention => C, Entity => Small_Identifier_Type);

   No_Small_Identifier : constant Small_Identifier_Type := 0;

   -- ------------------------------------------------------------------------

   type Identifier_Type is mod 2 ** Medium_Size;
   for Identifier_Type'Size use Medium_Size;
   pragma Convention (Convention => C, Entity => Identifier_Type);
   Null_Identifier : constant Identifier_Type := 0;

   -- ------------------------------------------------------------------------

   type Natural_Identifier_Type is mod 2 ** (Medium_Size - 1);
   for Natural_Identifier_Type'Size use Medium_Size;
   pragma Convention (Convention => C, Entity => Natural_Identifier_Type);
   No_Natural_Identifier : constant Natural_Identifier_Type := 0;

   -- ------------------------------------------------------------------------

   type Handle_Type is mod 2 ** Big_Size;
   for Handle_Type'Size use Big_Size;
   pragma Convention (Convention => C, Entity => Handle_Type);
   Null_Handle : constant Handle_Type := 0;
   Invalid_Handle : constant Handle_Type := (2 ** Handle_Type'Size) - 1;

   -- ------------------------------------------------------------------------

   type Padding_Type is (Padding_Value);
   for Padding_Type use (Padding_Value => 0);

   -- ------------------------------------------------------------------------

   type Reserved_Type is (Reserved_Value);
   for Reserved_Type use (Reserved_Value => 0);

   -- ------------------------------------------------------------------------

   type Undefined_Type is (Undefined);
   for Undefined_Type use (Undefined => 0);

   -- ------------------------------------------------------------------------

   type Variant_Kinds is
     (As_Access,
      As_Boolean,
      As_Coordinate,
      As_Handle,
      As_Signed,
      As_Size,
      As_Small_Identifiers,
      As_Unsigned);

   type Padding_For_Medium_Sized_Type is new Padding_Type;
   for Padding_For_Medium_Sized_Type'Size use
     (Access_Type'Size - Unsigned_Type'Size);

   type Variant_Type (Kind : Variant_Kinds := As_Access) is record
      case Kind is
         when As_Access =>
            Pointer : Access_Type;
         when As_Boolean =>
            Status : Boolean_Type;
         when As_Coordinate =>
            X : Small_Signed_Type; -- Low
            Y : Small_Signed_Type; -- High
         when As_Handle =>
            Handle : Handle_Type;
         when As_Signed =>
            Signed         : Signed_Type;
            Signed_Padding : Padding_For_Medium_Sized_Type := Padding_Value;
         when As_Size =>
            Width  : Small_Signed_Type; -- Low
            Height : Small_Signed_Type; -- High
         when As_Small_Identifiers =>
            Identifier_1 : Small_Identifier_Type; -- Low word
            Identifier_2 : Small_Identifier_Type; -- High word
         when As_Unsigned =>
            Unsigned         : Signed_Type;
            Unsigned_Padding : Padding_For_Medium_Sized_Type := Padding_Value;
      end case;
   end record;

   pragma Unchecked_Union (Variant_Type);
   pragma Pack (Variant_Type);
   for Variant_Type'Size use Access_Type'Size;
   pragma Convention (Convention => C_Pass_By_Copy, Entity => Variant_Type);

   Null_Variant : constant Variant_Type :=
     (Kind    => As_Access,
      Pointer => Null_Access);

   -- ------------------------------------------------------------------------

   type Padding_For_Small_Sized_Type is new Padding_Type;
   for Padding_For_Small_Sized_Type'Size use
     (Access_Type'Size - Small_Identifier_Type'Size);

   type Variant_Identifier_Kinds is (By_Number, By_Name);

   type Identifier_Name_Type is
     array (Positive range <>) of aliased API_Character_Type;

   pragma Convention (Convention => C, Entity => Identifier_Name_Type);

   type Identifier_Name_Access is access constant Identifier_Name_Type;
   for Identifier_Name_Access'Size use Access_Type'Size;
   pragma Convention (Convention => C, Entity => Identifier_Name_Access);

   type Variant_Identifier_Type
     (Kind : Variant_Identifier_Kinds := By_Number)
   is
      record
         case Kind is
            when By_Number =>
               Number  : Small_Identifier_Type := 0;
               Padding : Padding_For_Small_Sized_Type := Padding_Value;
            when By_Name =>
               Name : Identifier_Name_Access;
         end case;
      end record;

   pragma Unchecked_Union (Variant_Identifier_Type);
   pragma Pack (Variant_Identifier_Type);
   for Variant_Identifier_Type'Size use Access_Type'Size;
   pragma Convention
     (Convention => C_Pass_By_Copy,
      Entity     => Variant_Identifier_Type);

   Null_Variant_Identifier : constant Variant_Identifier_Type :=
     (Kind    => By_Number,
      Number  => 0,
      Padding => <>);

   -- ------------------------------------------------------------------------

   type Two_Small_Unsigned_Type is record
      Unsigned_1 : Small_Unsigned_Type;
      Unsigned_2 : Small_Unsigned_Type;
   end record;

   for Two_Small_Unsigned_Type use record
      Unsigned_1 at 0 range
         0 .. (Small_Unsigned_Type'Size - 1);
      Unsigned_2 at 0 range
         Small_Unsigned_Type'Size .. (Unsigned_Type'Size - 1);
   end record;

   for Two_Small_Unsigned_Type'Size use Unsigned_Type'Size;
   pragma Convention
     (Convention => C_Pass_By_Copy,
      Entity     => Two_Small_Unsigned_Type);

   -- ------------------------------------------------------------------------

   procedure Set_Last_Error (Error_Code : Error_Code_Type);

   pragma Import
     (Convention => Stdcall,
      Entity => Set_Last_Error,
      External_Name => "SetLastError");

   for Error_Code_Type'Size use Medium_Size;
   pragma Convention (Convention => C, Entity => Error_Code_Type);

   No_Error : constant Error_Code_Type := 0;

   for Exit_Code_Type'Size use Medium_Size;
   pragma Convention (Convention => C, Entity => Exit_Code_Type);

   Terminates_Successfully : constant Exit_Code_Type := 0;

   -- ------------------------------------------------------------------------

   package Messages is

      -- ---------------------------------------------------------------------

      subtype Message_Target_Type is Handle_Type;

      No_Target : constant Message_Target_Type :=
         Message_Target_Type (Null_Handle);

      Thread_Target : constant Message_Target_Type := Null_Handle;

      Broadcast_Target : constant Message_Target_Type := 16#FFFF#;

      type Target_Filter_Type is new Message_Target_Type;

      Any_Target : constant Target_Filter_Type :=
         Target_Filter_Type (Null_Handle);

      -- ---------------------------------------------------------------------

      type Message_Identifier_Type is new Identifier_Type;

      -- ---------------------------------------------------------------------

      type Message_Occurence_Time_Type is new Unsigned_Type;

      type Message_Mouse_Coordinate_Type is new Signed_Type range
         -2_147_483_648 .. +2_147_483_647;

      type Message_Mouse_Position_Type is record
         X : Message_Mouse_Coordinate_Type;
         Y : Message_Mouse_Coordinate_Type;
      end record;

      pragma Pack (Message_Mouse_Position_Type);
      for Message_Mouse_Position_Type'Size use 64;
      pragma Convention
        (Convention => C_Pass_By_Copy,
         Entity     => Message_Mouse_Position_Type);

      -- ---------------------------------------------------------------------

      type Message_Type is record
         Target : Message_Target_Type;
         Message_Identifier : Message_Identifier_Type;
         Parameter_1 : Variant_Type;
         Parameter_2 : Variant_Type;
         Occurence_Time : Message_Occurence_Time_Type;
         Mouse_Position : Message_Mouse_Position_Type;
      end record;

      pragma Pack (Message_Type);
      for Message_Type'Size use 224;
      pragma Convention
        (Convention => C_Pass_By_Copy,
         Entity     => Message_Type);

      -- ---------------------------------------------------------------------

      function Send_Message
        (Target             : Message_Target_Type;
         Message_Identifier : Message_Identifier_Type;
         Parameter_1        : Variant_Type;
         Parameter_2        : Variant_Type)
         return               Variant_Type;

      pragma Import
        (Convention    => Stdcall,
         Entity        => Send_Message,
         External_Name => "SendMessageW");

      -- ---------------------------------------------------------------------

      function Post_Message
        (Target             : Message_Target_Type;
         Message_Identifier : Message_Identifier_Type;
         Parameter_1        : Variant_Type;
         Parameter_2        : Variant_Type)
         return               Variant_Type;

      pragma Import
        (Convention    => Stdcall,
         Entity        => Post_Message,
         External_Name => "PostMessageW");

      -- ---------------------------------------------------------------------

      function Get_Message
        (Message                  : not null access Message_Type;
         Target                   : Target_Filter_Type      := Any_Target;
         First_Message_Identifier : Message_Identifier_Type := 0;
         Last_Message_Identifier  : Message_Identifier_Type := 0)
         return                     Boolean_Type;

      pragma Import
        (Convention    => Stdcall,
         Entity        => Get_Message,
         External_Name => "GetMessageW");

      -- ---------------------------------------------------------------------

      type Remove_Behaviour_Type is new Options_Type;

      Dont_Remove : constant Remove_Behaviour_Type := 16#00000000#;
      Always_Remove : constant Remove_Behaviour_Type := 16#00000001#;
      Dont_Yield : constant Remove_Behaviour_Type := 16#00000002#;
      Process_Paint : constant Remove_Behaviour_Type := 16#00200000#;
      Process_Sended_Messages : constant Remove_Behaviour_Type :=
         16#00400000#;
      Process_Mouse_And_Keyboard_Input : constant Remove_Behaviour_Type :=
         16#00070000#;
      Process_Posted_Messages : constant Remove_Behaviour_Type :=
         16#00980000#;

      -- ---------------------------------------------------------------------

      function Peek_Message
        (Message                  : not null access Message_Type;
         Target                   : Target_Filter_Type      := Any_Target;
         First_Message_Identifier : Message_Identifier_Type := 0;
         Last_Message_Identifier  : Message_Identifier_Type := 0;
         Remove_Behaviour         : Remove_Behaviour_Type   := Dont_Remove)
         return                     Boolean_Type;

      pragma Import
        (Convention    => Stdcall,
         Entity        => Peek_Message,
         External_Name => "PeekMessageW");

      -- ---------------------------------------------------------------------

      Current_Dialog : Handle_Type := Null_Handle;

      function Handle_Dialog_Message
        (Window  : Handle_Type;
         Message : not null access Message_Type)
         return    Boolean_Type;

      pragma Import
        (Convention    => Stdcall,
         Entity        => Handle_Dialog_Message,
         External_Name => "IsDialogMessageW");

      -- ---------------------------------------------------------------------

      procedure Translate_Message (Message : not null access Message_Type);

      pragma Import
        (Convention    => Stdcall,
         Entity        => Translate_Message,
         External_Name => "TranslateMessage");

      -- ---------------------------------------------------------------------

      procedure Dispatch_Message (Message : not null access Message_Type);

      pragma Import
        (Convention    => Stdcall,
         Entity        => Dispatch_Message,
         External_Name => "DispatchMessageW");

      -- ---------------------------------------------------------------------

      procedure Post_Quit_Message (Exit_Code : Exit_Code_Type);

      pragma Import
        (Convention    => Stdcall,
         Entity        => Post_Quit_Message,
         External_Name => "PostQuitMessage");

   end Messages;

   -- ------------------------------------------------------------------------

   package Behaviours is

      type Event_Data_Type is record
         Message_Identifier : Messages.Message_Identifier_Type;
         Target_Object      : Messages.Message_Target_Type;
         Parameter_1        : Variant_Type;
         Parameter_2        : Variant_Type;
      end record;

      type Method_Type is access procedure
        (Event_Data : in Event_Data_Type);

      type Agent_Type is tagged null record;

      procedure Execute (Agent : in Agent_Type);

      type Agent_Access is access constant Agent_Type'Class;

      type Behaviours_Map_Type (<>) is limited private;

      function New_Behaviours_Map return Behaviours_Map_Type;

      -- ---------------------------------------------------------------------

      procedure Associate
        (Map                : in out Behaviours_Map_Type;
         Agent              : in Agent_Access;
         Message_Identifier : in Messages.Message_Identifier_Type);

      procedure Associate
        (Map                : in out Behaviours_Map_Type;
         Method             : in Method_Type;
         Message_Identifier : in Messages.Message_Identifier_Type);

      -- ---------------------------------------------------------------------

      procedure Associate
        (Map                : in out Behaviours_Map_Type;
         Agent              : in Agent_Access;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Target_Object      : in Messages.Message_Target_Type);

      procedure Associate
        (Map                : in out Behaviours_Map_Type;
         Method             : in Method_Type;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Target_Object      : in Messages.Message_Target_Type);

      -- ---------------------------------------------------------------------

      procedure Associate
        (Map                : in out Behaviours_Map_Type;
         Agent              : in Agent_Access;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Target_Identifier  : in Small_Identifier_Type);

      procedure Associate
        (Map                : in out Behaviours_Map_Type;
         Method             : in Method_Type;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Target_Identifier  : in Small_Identifier_Type);

      -- ---------------------------------------------------------------------

      procedure Execute
        (Map                : in out Behaviours_Map_Type;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Parameter_1        : in Variant_Type;
         Parameter_2        : in Variant_Type);

      procedure Execute
        (Map                : in out Behaviours_Map_Type;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Target_Object      : in Messages.Message_Target_Type;
         Parameter_1        : in Variant_Type;
         Parameter_2        : in Variant_Type);

      procedure Execute
        (Map                : in out Behaviours_Map_Type;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Target_Identifier  : in Small_Identifier_Type;
         Parameter_1        : in Variant_Type;
         Parameter_2        : in Variant_Type);

      procedure Execute
        (Map                : in out Behaviours_Map_Type;
         Message_Identifier : in Messages.Message_Identifier_Type;
         Target_Object      : in Messages.Message_Target_Type;
         Target_Identifier  : in Small_Identifier_Type;
         Parameter_1        : in Variant_Type;
         Parameter_2        : in Variant_Type);

   private

      subtype Message_Identifier_Type is Messages.Message_Identifier_Type;

      type Association_Type;

      type Association_Access is access Association_Type;

      type Behaviours_Map_Type is record
         First : Association_Access;
         Last  : Association_Access;
      end record;

   end Behaviours;

   -- ------------------------------------------------------------------------

   generic

      type Parameters_Type is private;
      with procedure Execute (Parameters : Parameters_Type);

   package Agents is

      type Agent_Type is new Behaviours.Agent_Type with private;

      function New_Agent (Parameters : Parameters_Type) return Agent_Type;

   private

      type Agent_Type is new Behaviours.Agent_Type with record
         Parameters : Parameters_Type;
      end record;

      pragma Warnings (Off);

      overriding procedure Execute (Agent : in Agent_Type);

      pragma Warnings (On);

   end Agents;

   -- ------------------------------------------------------------------------

   package Parameterless_Agents is

      type Method_Access is access procedure;

      type Agent_Type is new Behaviours.Agent_Type with private;

      function New_Agent (Method : Method_Access) return Agent_Type;

   private

      type Agent_Type is new Behaviours.Agent_Type with record
         Method : Method_Access;
      end record;

      pragma Warnings (Off);

      overriding procedure Execute (Agent : in Agent_Type);

      pragma Warnings (On);

   end Parameterless_Agents;

end Windows;
